/*
 * Created on 11.05.2005
 *
 */
package org.eclipse.contribution.junit;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.ServerSocket;
import java.net.Socket;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.debug.core.DebugEvent;
import org.eclipse.debug.core.IDebugEventSetListener;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.jdt.core.IJavaProject;

/**
 * @author Andreas
 *
 */
public class Communicate extends Job implements Runnable, IDebugEventSetListener{
	int port = 7777;
	BufferedReader reader;
	IJavaProject project;
	ILaunch launch;
	String failedtest;
	static Communicate instance;
	ServerSocket server;
	Socket socket;
	private boolean working;
	
	/**
	 * @param project
	 */
	public Communicate(IJavaProject project) {
		super("JUnit Plugin / Communicate");
		this.project = project;
	}

	public static Communicate getInstance(IJavaProject project){
		if (instance == null)
			instance = new Communicate(project);
		return instance;
	}
	
	public void setPort(int port){
		this.port = port;
	}
	
	/**
	 * overrides method belongsTo in Type Job.
	 */
	public boolean belongsTo(Object family) {
        return family == JUnitPlugin.JOBFAMILY;
     }
	
	public void run(){
		try {
			server = new ServerSocket(port);
			try {
				socket = server.accept();
				try {					
					readMessage();
				} finally {
					socket.close();
				}
			} finally {
				server.close();
			}
		} catch (IOException e){
			e.printStackTrace();
		}
	}
	
	private void readMessage() throws IOException{
		reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		try {
			working = true;
			String line = null;
			while ((line = reader.readLine()) != null && working){
				parseMessage(line);
			}
			working = false;
		}catch (Exception e){
			//fallthorugh;
		}finally {
			reader.close();
		}
	}
	
	private void parseMessage(String line){
		JUnitPlugin plugin = JUnitPlugin.getPlugin();
		if (line.startsWith("starting tests ")){
			int start = "starting tests ".length();
			int count = Integer.parseInt(line.substring(start));
			plugin.fireTestsStarted(count, project);
			if (JUnitPlugin.trace)
				System.out.println("TestS started:");
		}
		
		if (line.startsWith("ending test ")){
			int start = "ending test ".length();
			String method = line.substring(start, line.indexOf("("));
			String klass = line.substring(line.indexOf("(") + 1, line.indexOf(")"));
			//test already failed!
			if (failedtest != null && failedtest.equalsIgnoreCase(klass.toString()+" "+method.toString())){
				failedtest = null;
				return;
			}
			plugin.fireTestOk(klass, method, project);
			if (JUnitPlugin.trace)
				System.out.println("Test Ok: "+klass.toString()+" "+method.toString());
		}
		
		if (line.startsWith("ending tests")){
			plugin.fireTestsFinished(project);
			if (JUnitPlugin.trace)
				System.out.println("TestS ended:");
		}
		
		if (line.startsWith("starting test ")){
			int start = "starting test ".length();
			String method = line.substring(start, line.indexOf("("));
			String klass = line.substring(line.indexOf("(") + 1, line.indexOf(")"));
			plugin.fireTestStarted(klass, method, project);
			if (JUnitPlugin.trace)
				System.out.println("Test started: "+klass.toString()+" "+method.toString());
		}
		
		if (line.startsWith("failing test ")){
			int start = "failing test ".length();
			String method = line.substring(start, line.indexOf("("));
			String klass = line.substring(line.indexOf("(") + 1, line.indexOf(")"));
			StringWriter buffer = new StringWriter();
			PrintWriter writer = new PrintWriter(buffer);
			String frame = null;
			try {
				while ((frame = reader.readLine()) != null && (!frame.equals("END TRACE")))
					writer.println(frame);
			} catch (IOException e){
				e.printStackTrace();
			}
			String trace = buffer.getBuffer().toString();
			plugin.fireTestFailed(klass, method, trace, project);
			failedtest = klass.toString()+" "+method.toString();
			if (JUnitPlugin.trace)
				System.out.println("Test failed: "+klass.toString()+" "+method.toString());
		}
	}

	/**
	 * @param launch
	 */
	public void setLaunch(ILaunch launch) {
		this.launch = launch;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.core.runtime.jobs.Job#run(org.eclipse.core.runtime.IProgressMonitor)
	 */
	protected IStatus run(IProgressMonitor monitor) {
		try{
			run();
		}catch(Exception e){
			e.printStackTrace();
		}
		return Status.OK_STATUS;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.debug.core.IDebugEventSetListener#handleDebugEvents(org.eclipse.debug.core.DebugEvent[])
	 */
	public void handleDebugEvents(DebugEvent[] events) {
		try{
		if (events.length!= 0){
			//System.out.println(events[0].toString());
			if ((events[0].getKind() == DebugEvent.SUSPEND && (events[0].getDetail() == DebugEvent.STEP_END ||
					events[0].getDetail() == DebugEvent.BREAKPOINT))){
				NavigateDebug.getInstance().frameSet();
			}
		}
		}catch(Exception e){
			e.printStackTrace();
		}
	}

	/**
	 * 
	 */
	public void cancelJob() {
		working = false;
	}

}
