/*
 * Created on 06.04.2005
 *
 */
package org.eclipse.contribution.junit;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;

/**
 * @author Andreas
 *
 */
public class MarkerCreator implements ITestRunListener {
	public static final String OK_MARKER = "org.eclipse.contribution.junit.okMarker"; 
	public static final String FAILURE_MARKER = "org.eclipse.contribution.junit.failureMarker"; 
	public static final String UNKNOWN_MARKER = "org.eclipse.contribution.junit.unknownMarker"; 
	public static final String CHANGED_MARKER = "org.eclipse.contribution.junit.changedMarker";
		
	public static final String int_     = "I";
	public static final String short_   = "S";
	public static final String byte_    = "B";
	public static final String long_    = "J";
	public static final String double_  = "D";
	public static final String float_   = "F";
	public static final String boolean_ = "Z";
	public static final String char_    = "C";

	
	private static MarkerCreator instance;
	
	/**
	 * 
	 */
	public MarkerCreator() {
		instance = this;
	}

	public static MarkerCreator getInstance(){
		if( instance == null )
			instance = new MarkerCreator();
		return instance;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testsStarted(int, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testsStarted(int testCount, IJavaProject project) {
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testsFinished(org.eclipse.jdt.core.IJavaProject)
	 */
	public void testsFinished(IJavaProject project) {
	}


	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testStarted(java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testStarted(String klass, String method, IJavaProject project) {
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testOk(java.lang.String, java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testOk(String klass, String method, IJavaProject project) {
		ExtendedTestResult testresult = (ExtendedTestResult) NavigateDebug.getInstance().tests.getLast();

		try {
			createMarker(klass, method, "test ok!", project, OK_MARKER, testresult);
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testFailed(java.lang.String, java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testFailed(String klass, final String method, final String trace,
			IJavaProject project) {
		ExtendedTestResult testresult = (ExtendedTestResult) NavigateDebug.getInstance().tests.getLast();
		
		try {
			createMarker(klass, method, trace, project, FAILURE_MARKER, testresult);
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Creates the colored markers.
	 * @param klass
	 * @param method
	 * @param trace
	 * @param testProject
	 * @param markerType
	 * @param testresult
	 * @throws CoreException
	 * @throws JavaModelException
	 */
	
	private void createMarker(final String klass, final String method, final String trace, IJavaProject testProject, final String markerType, ExtendedTestResult testresult) throws CoreException, JavaModelException{
		String testedklass;
		TestedMethod testedmethod;
		String testmethod;
		LinkedList testedmethods;
		String [] args;
		Iterator iter;
		Iterator number;
		Object o;
		ArrayList linenumbers;
		IMarker[] dirtymarkers;
		int tempoffset;
		IType type;
		ISourceRange sr;
		boolean markercreation;
		
		
//		Class and Method doesn't match
		if (!(klass.equalsIgnoreCase(testresult.getKlass()) && method.equalsIgnoreCase(testresult.getMethod())))
			return;
		
		type = null;
		
		testedmethods = testresult.testedMethods;
		iter = testedmethods.iterator();
		//for all tested methods
		while (iter.hasNext()){
			o = iter.next();
			if (o instanceof TestedMethod){
				testedmethod = (TestedMethod) o;
				testedklass = testedmethod.getKlass();
				testmethod = testedmethod.getMethod();
				linenumbers = testedmethod.linenumbers;
							
				type = testProject.findType(testedklass);
				if (type == null) return;
				final IResource resource = type.getUnderlyingResource();
				
				args = testedmethod.getArgsAsString();
				for (int i = 0; i<args.length;i++){
					args[i] = mapType(args[i]);
				}

				//Class, Constructor			
				if (testmethod.startsWith("<")){
					sr = null;
				} else{
					sr = type.getMethod(testmethod, args).getSourceRange();
				}
				
				final ISourceRange range = sr;
				
				dirtymarkers = resource.findMarkers("org.eclipse.contribution.junit.dirtyMarker", true, IResource.DEPTH_INFINITE);
				markercreation = true;
				for (int i=0; i<dirtymarkers.length;i++){
					//method is dirty, no need to create markers
					tempoffset = ((Integer) dirtymarkers[i].getAttribute("org.eclipse.contribution.junit.offset")).intValue();
					if (range != null && tempoffset >= range.getOffset() &&
						tempoffset <= range.getOffset()+range.getLength()){
							markercreation = false;
					}
				}
				
				number = linenumbers.iterator();
				
				//for each linenumber, a colored marker is created.
				while (markercreation && number.hasNext()){
					o = number.next();
					if (o instanceof Integer){
						int temp = ((Integer) o).intValue();
						temp = temp + changeOffset(resource, temp);
						final Integer i = new Integer(temp);
						IWorkspaceRunnable runnable = new IWorkspaceRunnable(){
							public synchronized void run(IProgressMonitor monitor) throws CoreException {
								IMarker marker = resource.createMarker(JUnitPlugin.TEST_MARKER);
								Map map = new HashMap(6);
								if (range != null){
									map.put("org.eclipse.contribution.junit.charstart", new Integer(range.getOffset()));
									map.put("org.eclipse.contribution.junit.charend", new Integer(range.getOffset()+range.getLength()));
								}
								map.put(IMarker.MESSAGE, trace);
								map.put(IMarker.LINE_NUMBER, i);
								map.put("org.eclipse.contribution.junit.testclass", klass);
								map.put("org.eclipse.contribution.junit.testmethod", method);
								marker.setAttributes(map);
								marker.setAttribute("org.eclipse.contribution.junit.type", checkMarkerType(resource, i, markerType));
							}
						};
						resource.getWorkspace().run(runnable, null);
					}
				}
			}
		}
	}
	
	
	/**
	 * If the user changes any lines and the plugin is still running, all linenumbers
	 * below the change will be new calculated
	 * 
	 * @param resource
	 * @param i2
	 * @return
	 */
	
	private int changeOffset(IResource resource, int linenumber) {
		IMarker [] linechangemarkers;
		int retval = 0;
		Integer markeroffset;
		Integer changedlinenumber;
		
		try {
			if (resource == null)
				return 0;
			linechangemarkers = resource.findMarkers("org.eclipse.contribution.junit.lineChangeMarker", true, IResource.DEPTH_INFINITE);
			for (int i=0; i<linechangemarkers.length; i++){
				markeroffset = (Integer) linechangemarkers[i].getAttribute("org.eclipse.contribution.junit.offset");
				changedlinenumber = (Integer) linechangemarkers[i].getAttribute("org.eclipse.contribution.junit.changedlinenumber");
				if (changedlinenumber.intValue()+2 < linenumber){
					retval = retval + markeroffset.intValue();
				}
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		}
		return retval;
	}
	
	
	private String getArray(String s, int index){
		String val = "";
		String arr = "";
		
		if (index == -1)
			return arr;
		
		val = s.substring(index);
		while (!val.equals("") && val.charAt(0)=='['){
			val = val.substring(2);
			arr = arr+"[";
		}
		return arr;
	}
	
	/**
	 * Maps the type of a specified methodargument.
	 * 
	 * @param s
	 * @return
	 */
	private String mapType(String s){
		String val="";
		String temp="";
		
		if (s.startsWith("int")){
			return getArray(s, s.indexOf('['))+int_;
		} else if (s.startsWith("short")){
			return getArray(s, s.indexOf('['))+short_;
		} else if (s.startsWith("byte")){
			return getArray(s, s.indexOf('['))+byte_;
		} else if (s.startsWith("long")){
			return getArray(s, s.indexOf('['))+long_;
		} else if (s.startsWith("double")){
			return getArray(s, s.indexOf('['))+double_;
		} else if (s.startsWith("float")){
			return getArray(s, s.indexOf('['))+float_;
		} else if (s.startsWith("boolean")){
			return getArray(s, s.indexOf('['))+boolean_;
		} else if (s.startsWith("char")){
			return getArray(s, s.indexOf('['))+char_;
		} else {
			val = s.substring(s.lastIndexOf('.')+1); //deletes packagenames
			temp = val;
			if (val.indexOf('[')!= -1)
				temp = val.substring(0, val.indexOf('['));
			return getArray(val, val.indexOf('['))+"Q"+temp+";";
		}
	}

	
	/**
	 * calculates the needed color for the given markerType
	 * 
	 * @param resource
	 * @param i
	 * @param markertype
	 * @param l
	 * @return the calculatet markertype
	 */
	private String checkMarkerType(IResource resource, Integer i, String markertype) {
		try {
			IMarker [] markers = resource.findMarkers(JUnitPlugin.TEST_MARKER, false, IResource.DEPTH_ZERO);
			for (int j = 0; j<markers.length; j++){
				if (markers[j].getAttribute(IMarker.LINE_NUMBER, -1) == i.intValue()){
					if (markers[j].getAttribute("org.eclipse.contribution.junit.type", "testMarker")
							.equalsIgnoreCase(OK_MARKER)&&(markertype != OK_MARKER)){
						markers[j].setAttribute("org.eclipse.contribution.junit.type", UNKNOWN_MARKER);
						return UNKNOWN_MARKER;
					}else if (markers[j].getAttribute("org.eclipse.contribution.junit.type", "testMarker")
							.equalsIgnoreCase(FAILURE_MARKER)&&(markertype != FAILURE_MARKER)){
						markers[j].setAttribute("org.eclipse.contribution.junit.type", UNKNOWN_MARKER);
						return UNKNOWN_MARKER;
					}else if (markers[j].getAttribute("org.eclipse.contribution.junit.type", "testMarker")
							.equalsIgnoreCase(UNKNOWN_MARKER)){
						return UNKNOWN_MARKER;
					}
				}
			}
		} catch (CoreException e) {
			e.printStackTrace();
		}
		return markertype;
	}
}
