/*
 * Created on 16.03.2005
 *
 */
package org.eclipse.contribution.junit;

import java.io.IOException;
import java.net.URL;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.debug.core.DebugEvent;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.IDebugEventSetListener;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.ILaunchManager;
import org.eclipse.debug.core.Launch;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.jdt.launching.IVMRunner;
import org.eclipse.jdt.launching.JavaRuntime;
import org.eclipse.jdt.launching.SocketUtil;
import org.eclipse.jdt.launching.VMRunnerConfiguration;
import org.osgi.framework.Bundle;
/**
 * @author Andreas
 *
 */
public class TestRunner implements IDebugEventSetListener{
	
	static final String MAIN_CLASS = "org.eclipse.contribution.junit.SocketTestRunner";
	
	private int port;
	private IJavaProject project;
	public static TestRunner instance;
	
	public static TestRunner getInstance(IJavaProject project){
		if( instance == null )
		      instance = new TestRunner(project);
		return instance;
	}
	
	public TestRunner(IJavaProject project){
		this.project = project;
		instance = this;
	}	
	
	/**
	 * @param navdeb
	 * @param project2
	 * @param types
	 */
	public void run(IType[] classes, NavigateDebug navdeb) throws CoreException {
		IVMInstall vmInstall = getVMInstall();		
		if (vmInstall == null)
			return;
		IVMRunner vmRunner = vmInstall.getVMRunner(ILaunchManager.DEBUG_MODE);
		if (vmRunner == null)
			return;
		String[] classPath = computeClasspath();
		
		VMRunnerConfiguration vmConfig = new VMRunnerConfiguration(MAIN_CLASS, classPath);
		String[] args = new String[classes.length +1];
		port = SocketUtil.findFreePort();
				
		args[0] = Integer.toString(port);
		
		for (int i = 0; i < classes.length; i++)
			args[i+1] = classes[i].getFullyQualifiedName();
		vmConfig.setProgramArguments(args);
		vmConfig.setResumeOnStartup(false);
		
		ILaunch launch = new Launch(null, ILaunchManager.DEBUG_MODE, null);
		
		Communicate c = Communicate.getInstance(project);
		c.setPort(port);
		DebugPlugin.getDefault().addDebugEventListener(c);
		c.setLaunch(launch);
		c.schedule();
		
		vmRunner.run(vmConfig, launch, null);

		DebugPlugin.getDefault().getLaunchManager().addLaunch(launch);

		navdeb.setProject(project);
		navdeb.setMainThread(launch);
		navdeb.setLaunch(launch);
		navdeb.schedule();
	}

	private String[] computeClasspath() throws CoreException{
		String[] defaultPath = JavaRuntime.computeDefaultRuntimeClassPath(project);
		String[] classPath = new String[defaultPath.length + 2];
		System.arraycopy(defaultPath, 0, classPath, 2, defaultPath.length);
		JUnitPlugin plugin = JUnitPlugin.getPlugin();
		Bundle bundle = plugin.getBundle();
		URL url = bundle.getEntry("/");
		try{
			classPath[0] = Platform.asLocalURL(new URL(url, "bin")).getFile();
//			classPath[1] = Platform.asLocalURL(new URL(url, "sswjunit.jar")).getFile();
		} catch(IOException e){
			IStatus status = new Status(IStatus.ERROR, bundle.getSymbolicName(), IStatus.OK, "Could not determine path", e);
			throw new CoreException(status);
		}
		return classPath;
	}
	
	private IVMInstall getVMInstall() throws CoreException{
		IVMInstall vmInstall = JavaRuntime.getVMInstall(project);
		if (vmInstall == null)
			vmInstall = JavaRuntime.getDefaultVMInstall();
		return vmInstall;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.debug.core.IDebugEventSetListener#handleDebugEvents(org.eclipse.debug.core.DebugEvent[])
	 */
	public void handleDebugEvents(DebugEvent[] events) {
		if (events != null && events[0]!= null){
			if (events[0].getKind() == DebugEvent.SUSPEND && events[0].getDetail() == DebugEvent.STEP_END){
				NavigateDebug.getInstance().notify();
			}
		}
	}
}
