/*
 * Created on 20.04.2005
 *
 */
package org.eclipse.contribution.junit;

import java.util.*;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.*;
import org.eclipse.jdt.core.search.*;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.PlatformUI;

/**
 * @author Andreas
 *
 */
public class ClassSearcher {

	public IType[] findAllTestCases(IJavaProject project, IProgressMonitor pm)throws JavaModelException{
		IType[] candidates = getAllSubClasses(project, pm, "junit.framework.TestCase");
		return collectTestsInProject(candidates, project);
	}
	
	public IJavaElement[] findAllClassesExceptTestCases(IJavaProject project, IProgressMonitor pm)throws JavaModelException{
		ArrayList result = new ArrayList();
		IJavaElement[] classes = findAllClasses(project, pm);
		IType[] testcases = findAllTestCases(project, pm);
		for (int i = 0; i< classes.length; i++){
			if (isNotATestCase(classes[i], testcases))
				result.add(classes[i]);
		}
		return (IJavaElement[]) result.toArray(new IJavaElement[result.size()]); 	
	}
	
	/**
	 * @param element
	 * @param testcases
	 * @return
	 */
	private boolean isNotATestCase(IJavaElement element, IType[] testcases) {
		for (int i=0; i<testcases.length;i++){
			if (element.getElementName().equals(testcases[i].getElementName()+".java"))
				return false;
		}
		return true;
	}

	public ICompilationUnit[] findAllClasses(IJavaProject project, IProgressMonitor pm)throws JavaModelException{
		ArrayList files = new ArrayList();
		IPackageFragment [] packfrag = project.getPackageFragments();
		
		for (int i = 0; i<packfrag.length;i++){
			ICompilationUnit [] compunit = packfrag[i].getCompilationUnits();
			for (int j = 0; j<compunit.length;j++)
				files.add(compunit[j]);
		}
		ICompilationUnit [] classes = new ICompilationUnit[files.size()];
		files.toArray(classes);
		return classes;
	}
	
	
	public IJavaElement[] findAllOpenClasses(IJavaProject project, IProgressMonitor pm)throws JavaModelException{
		ArrayList openfiles = new ArrayList();
		Map map = new HashMap();
		IResource resource;
		String name;
		IPackageFragment [] packfrag = project.getPackageFragments();
		
		for (int i = 0; i<packfrag.length;i++){
			ICompilationUnit [] compunit = packfrag[i].getCompilationUnits();
			for (int j = 0; j<compunit.length;j++){
				resource = compunit[j].getResource();
				map.put(resource.getName(),compunit[j]);
			}
		}
		IEditorReference[] editref = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getEditorReferences();
		for (int i=0;i<editref.length;i++){
			name=editref[i].getName();
			if (map.containsKey(name))
				openfiles.add(map.get(name));
		}
		IJavaElement[] temp = new IJavaElement[openfiles.size()];
		openfiles.toArray(temp);
		return temp;
	}
	
	public void findMethodDeclaration(IJavaSearchScope scope, String methodname, String klassname){
		SearchPattern pattern = SearchPattern.createPattern(methodname,IJavaSearchConstants.METHOD,
				IJavaSearchConstants.DECLARATIONS,
				SearchPattern.R_EXACT_MATCH | SearchPattern.R_CASE_SENSITIVE ); 
		SearchEngine searchengine = new SearchEngine();
	    try {
	        SearchRequestor requestor = new JUnitSearchRequestor(klassname, methodname);
			searchengine.search(pattern, new SearchParticipant[] {SearchEngine.getDefaultSearchParticipant()}, scope, requestor, null);
			} catch (CoreException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param candidates
	 * @param project
	 * @return
	 */
	private IType[] collectTestsInProject(IType[] candidates, IJavaProject project) {
		List result = new ArrayList();
		for (int i = 0; i< candidates.length; i++){
			try{
				if (isTestInProject(candidates[i], project))
					result.add(candidates[i]);
			}catch (JavaModelException e){
				//Fall through
			}
		}
		return (IType[]) result.toArray(new IType[result.size()]); 
	}

	/**
	 * @param type
	 * @param project
	 * @return
	 */
	private boolean isTestInProject(IType type, IJavaProject project)throws JavaModelException {
		IResource resource = type.getUnderlyingResource();
		if (resource == null)
			return false;
		if (! resource.getProject().equals(project.getProject()))
			return false;
		return ! Flags.isAbstract(type.getFlags());
	}

	/**
	 * @param project
	 * @param pm
	 * @param klass 
	 * @return
	 */
	private IType[] getAllSubClasses(IJavaProject project, IProgressMonitor pm, String klass)throws JavaModelException {
		IType testCase = project.findType(klass);
		if (testCase == null)
			return new IType[0];
		ITypeHierarchy hierarchy = testCase.newTypeHierarchy(project, pm);
		return hierarchy.getAllSubtypes(testCase);
	}
}
