package org.eclipse.contribution.junit;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.DebugPlugin;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.texteditor.IDocumentProvider;

/**
 * @author Andreas
 *
 */
public class JUnitPlugin extends Plugin {
	
	private static JUnitPlugin instance;
	private List listeners = new ArrayList();
	public static final String AUTO_TEST_NATURE= "org.eclipse.contribution.junit.autoTestNature";
	public static final String AUTO_TEST_BUILDER= "org.eclipse.contribution.junit.autoTestBuilder";
	static boolean trace=false;
	static long starttime=0;
	static long endtime=0;
	public static final String JOBFAMILY = "debugFamily";
	public static boolean WORKING = false;
	public static Object lock = new Object();
	public Button button;
	
	public static final String TEST_MARKER = "org.eclipse.contribution.junit.testMarker";
	public static final String SEARCH_MARKER = "org.eclipse.contribution.junit.searchMarker";
	public static final String DIRTY_MARKER = "org.eclipse.contribution.junit.dirtyMarker";
	public static final String LINECHANGE_MARKER = "org.eclipse.contribution.junit.lineChangeMarker";
	public static final String TESTED_MARKER = "org.eclipse.contribution.junit.testedMarker";
	
		
	private IDocumentListener documentlistener;
	private ISelectionListener selectionlistener;
	private IWorkbenchPart actualpart;
	private IDocument actualdoc;
	
	IAction action;
	
	static{
		String value = Platform.getDebugOption("org.eclipse.contribution.junit/trace/testing");
		if (value != null && value.equalsIgnoreCase("true"))
			JUnitPlugin.trace = true;
	}
	
	public JUnitPlugin(){
		super();
		instance = this;
	}
	
	public static JUnitPlugin getPlugin(){
		return instance;
	}
	
	/**
	 * @param project
	 */
	
	public void cleanup(IJavaProject project){
		DebugPlugin.getDefault().removeDebugEventListener(Communicate.getInstance(project));
		removeTestListener(MarkerCreator.getInstance());
		removeTestListener(NavigateDebug.getInstance());
		removeSelectionListener();
		
		NavigateDebug.getInstance().cancelJob();
		
		//NavigateDebug could wait. frameSet() prevents a deadlock
		NavigateDebug.getInstance().frameSet();
		Communicate.getInstance(project).cancelJob();
		
		ILaunch [] launches = DebugPlugin.getDefault().getLaunchManager().getLaunches();
		for (int i = 0; i<launches.length; i++){
			try{
				launches[i].terminate();
			} catch (DebugException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * @param action
	 * @return
	 */
	public void run(IType[] types, IJavaProject project, IAction action) throws CoreException{
		this.action = action;
		if (action != null && !action.isChecked()){
			cleanup(project);
			return;
		}
		
		if (JUnitPlugin.trace){
			starttime = Calendar.getInstance().getTimeInMillis();
			System.out.println("STARTTIME: "+starttime);
		}
		
		IWorkbench workbench = PlatformUI.getWorkbench();
		
		//there are no tests in the project
		if (types == null || types.length == 0 || !workbench.saveAllEditors(false)){
			if (action != null)
				action.setChecked(false);
			return;
		}

		ITestRunListener listener1 = MarkerCreator.getInstance();
		NavigateDebug navdeb =  NavigateDebug.getInstance();
		TestRunner runner = TestRunner.getInstance(project);
		navdeb.tests = new LinkedList();
		
		cleanup(project);
		
		try {
			//Deletes markers
			project.getProject().deleteMarkers(TEST_MARKER, false, IResource.DEPTH_INFINITE);
			project.getProject().deleteMarkers(LINECHANGE_MARKER, true, IResource.DEPTH_INFINITE);
			project.getProject().deleteMarkers(DIRTY_MARKER, true, IResource.DEPTH_INFINITE);
		} catch (CoreException e2) {
			e2.printStackTrace();
		}
		
		DebugPlugin.getDefault().addDebugEventListener(Communicate.getInstance(project));
		
		addTestListener(navdeb);
		addTestListener(listener1);
		addSelectionListener();
		
		if (!WORKING)
			runner.run(types, navdeb);	
	}

	/**
	 * 
	 */
	private void addSelectionListener() {
		Display d = Display.getDefault();
		JUnitRunnable runnable = new JUnitRunnable();
		d.syncExec(runnable);
		IWorkbenchWindow window = runnable.getWindow();
		
		if (window == null)
			return;
		
		IWorkbenchPage page = window.getActivePage();
        final IEditorPart editorpart = page.getActiveEditor();
        actualpart = editorpart;
        addDocumentListener(actualpart);
        
        selectionlistener = new ISelectionListener(){
            public void selectionChanged(IWorkbenchPart part, ISelection selection){
            	removeDocumentListener(actualpart);
            	addDocumentListener(part);
            	actualpart = part;
            }
        };
        page.addSelectionListener(selectionlistener);
	}
	
	private void removeSelectionListener(){
		if (selectionlistener == null)
			return;
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
        page.removeSelectionListener(selectionlistener);
        if (actualpart != null)
        	removeDocumentListener(actualpart);
	}

	private void addDocumentListener(IWorkbenchPart part){
		if (!(part instanceof CompilationUnitEditor))
            return;
        CompilationUnitEditor editor = (CompilationUnitEditor)part;
        if (!(editor.getEditorInput() instanceof FileEditorInput))
        	return;
        FileEditorInput ei = (FileEditorInput)editor.getEditorInput();
        try {
        	//gets all Markers on the actual shown document in the editor
			IMarker [] testmarkers = ei.getFile().findMarkers(TEST_MARKER, true, IResource.DEPTH_ZERO);
			IMarker [] testedmarkers = ei.getFile().findMarkers(TEST_MARKER, true, IResource.DEPTH_ZERO);
	          
			
            IDocumentProvider dp = editor.getDocumentProvider();
            actualdoc = dp.getDocument(editor.getEditorInput());
            documentlistener = new JUnitDocumentListener(actualdoc, testmarkers, testedmarkers, ei.getFile());
                
            actualdoc.addDocumentListener(documentlistener);
		} catch (CoreException e) {
			e.printStackTrace();
		}
	}
	
	private void removeDocumentListener(IWorkbenchPart part){
		if ((documentlistener == null)||(actualdoc == null))
			return;
		actualdoc.removeDocumentListener(documentlistener);
	}
	
	/**
	 * @param listener
	 */
	public void addTestListener(ITestRunListener listener) {
		listeners.add(listener);
	}

	/**
	 * @param listener
	 */
	public void removeTestListener(ITestRunListener listener) {
		listeners.remove(listener);
	}

	/**
	 * @param count
	 * @param project 
	 */
	public void fireTestsStarted(int count, IJavaProject project) {
		for (Iterator all = listeners.iterator(); all.hasNext();){
			ITestRunListener each = (ITestRunListener) all.next();
			each.testsStarted(count, project);
		}
	}

	/**
	 * @param project
	 * 
	 */
	public void fireTestsFinished(IJavaProject project) {
		for (Iterator all = listeners.iterator(); all.hasNext();){
			ITestRunListener each = (ITestRunListener) all.next();
			each.testsFinished(project);
		}
		
		if (action != null)
			action.setChecked(false);
	}

	/**
	 * @param klass
	 * @param method
	 * @param project
	 */
	public void fireTestStarted(String klass, String method, IJavaProject project) {
		for (Iterator all = listeners.iterator(); all.hasNext();){
			ITestRunListener each = (ITestRunListener) all.next();
			each.testStarted(klass, method, project);
		}
	}
	
	/**
	 * @param klass
	 * @param method
	 * @param project
	 */
	public void fireTestOk(String klass, String method, IJavaProject project) {
		for (Iterator all = listeners.iterator(); all.hasNext();){
			ITestRunListener each = (ITestRunListener) all.next();
			each.testOk(klass, method, project);
		}
	}

	/**
	 * @param klass
	 * @param method
	 * @param trace
	 * @param project
	 */
	public void fireTestFailed(String klass, String method, String trace, IJavaProject project) {
		for (Iterator all = listeners.iterator(); all.hasNext();){
			ITestRunListener each = (ITestRunListener) all.next();
			each.testFailed(klass, method, trace, project);
		}
	}
	
	public void addAutoBuildNature(IProject project)throws CoreException{
		if (project.hasNature(AUTO_TEST_NATURE))
			return;
		
		IProjectDescription description = project.getDescription();
		String[] ids= description.getNatureIds();
		String[] newIds= new String[ids.length+1];
		System.arraycopy(ids, 0, newIds, 0, ids.length);
		newIds[ids.length]=AUTO_TEST_NATURE;
		description.setNatureIds(newIds);
		project.setDescription(description, null);
	}
	
	public void removeAutoBuildNature(IProject project) throws CoreException{
		IProjectDescription description = project.getDescription();
		String[] ids = description.getNatureIds();
		for (int i=0; i< ids.length; i++){
			if (ids[i].equals(AUTO_TEST_NATURE)){
				String[] newIds = new String[ids.length-1];
				System.arraycopy(ids, 0, newIds, 0, i);
				System.arraycopy(ids, i+1, newIds, i, ids.length-i-1);
				description.setNatureIds(newIds);
				project.setDescription(description, null);
				return;
			}
		}
	}
}
