/*
 * Created on 30.06.2005
 *
 */
package org.eclipse.contribution.junit;

import java.util.Calendar;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.ILaunch;
import org.eclipse.debug.core.model.IBreakpoint;
import org.eclipse.debug.core.model.IThread;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaLineBreakpoint;
import org.eclipse.jdt.debug.core.IJavaStackFrame;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.JDIDebugModel;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;

/**
 * @author Andreas
 *
 */
public class NavigateDebug extends Job implements ITestRunListener{

	public LinkedList tests;
	private ILaunch launch;
	private IJavaProject project;
	private static NavigateDebug instance;
	private IJavaThread mainthread;
	private boolean testsstartet = false;
	private boolean working = false;
		
	/**
	 * 
	 */
	public NavigateDebug() {
		super("JUnit Plugin / NavigateDebug");
		tests = new LinkedList();
		instance = this;
	}
	
	public static NavigateDebug getInstance(){
		if( instance == null )
		      instance = new NavigateDebug();
		return instance;
	}
	
	/**
	 * overrides method belongsTo in Type Job.
	 */
	public boolean belongsTo(Object family) {
        return family == JUnitPlugin.JOBFAMILY;
     }
	
	/* (non-Javadoc)
	 * @see org.eclipse.core.internal.jobs.InternalJob#cancel()
	 */
	public void cancelJob() {
		working = false;
	}
	
	public void setMainThread(ILaunch launch){
		IJavaDebugTarget dtarget = null;
		IThread [] threads;
		
		if (launch.getDebugTarget() != null && launch.getDebugTarget() instanceof IJavaDebugTarget)
			dtarget = (IJavaDebugTarget) launch.getDebugTarget();
		
		if (dtarget == null)
			return;
		
			try {
				threads = dtarget.getThreads();

				for (int i = 0; i < threads.length; i++)
					if ((threads[i].getName()).equals("main")
							&& threads[i] instanceof IJavaThread)
						this.mainthread = (IJavaThread) threads[i];
			} catch (Exception e) {
				if (JUnitPlugin.trace)
					e.printStackTrace();
			}
	}
	
	
	public synchronized void calculatePath() throws DebugException, JavaModelException, InterruptedException, com.sun.jdi.VMDisconnectedException{
		IJavaThread main;
		IJavaStackFrame javaframe;
		ClassSearcher cs = new ClassSearcher();
		IJavaElement[] allclasses;
		IJavaElement[] classesnotests;
		
		allclasses = cs.findAllClasses(project, null);
		classesnotests = cs.findAllClassesExceptTestCases(project, null);

		if (mainthread == null)
			return;
		
		main = mainthread;
		
		//the breakpoints are set, so the whole debugtarget can resume to
		//jump to the first linebreakpoint.
		launch.getDebugTarget().resume();
			
		this.wait();
		
		if (JUnitPlugin.trace){
			if (!main.isSuspended()){
				System.out.println("MAIN IS NOT SUSPENDED!");
				return;
			}
			System.out.println("MAIN IS SUSPENDED!");
		}else
			if (!main.isSuspended())
				return;	
		javaframe = (IJavaStackFrame) main.getTopStackFrame();
		
		working = true;
		
		//for socketsynchronization. SocketTestRunner is executing, but listener is not indicating "tests Started" - For DEBUG
		while (!testsstartet){};
		
		while(working && javaframe!=null){
			while (javaframe!=null && isDeclaringTypeInProject(allclasses, javaframe.getDeclaringTypeName())){
				try{
					appendMethodToTest(javaframe, (ExtendedTestResult) tests.getLast(), classesnotests);
				}catch(NoSuchElementException e){
					e.printStackTrace();
				}
				main.stepInto();
				this.wait();
				javaframe = (IJavaStackFrame) main.getTopStackFrame();
				if ((javaframe != null && javaframe.getDeclaringTypeName().equals("org.eclipse.contribution.junit.SocketTestRunner"))|| working == false){
					main.resume();
					return;
				}
			}
			main.stepReturn();
			this.wait();
			javaframe = (IJavaStackFrame) main.getTopStackFrame();
			if (javaframe != null && javaframe.getDeclaringTypeName().equals("org.eclipse.contribution.junit.SocketTestRunner")){
				main.resume();
				working = false;
				return;
			}
		}
		main.resume();
		working = false;
	}
	
	public synchronized void frameSet(){
		this.notifyAll();
	}
	
	/**
	 * 
	 */
	private void printLinkedList() {
		ExtendedTestResult etr;
		TestedMethod tm;
		Iterator iter2;
		
		Iterator iter1 = tests.iterator();
		while (iter1.hasNext()){
			etr = (ExtendedTestResult) iter1.next();
			System.out.println("TESTCLASS: "+etr.getKlass());
			System.out.println("TESTMETHOD: "+etr.getMethod());
			System.out.println("TESTRESULT: "+etr.getResult());
			iter2 = etr.testedMethods.iterator();
			while (iter2.hasNext()){
				tm = (TestedMethod) iter2.next();
				System.out.println(tm.toString());
			}
		}
	}

	/**
	 * @param javaframe
	 * @param last
	 * @param classesnotests
	 */
	private void appendMethodToTest(IJavaStackFrame javaframe, ExtendedTestResult test, IJavaElement[] classesnotests) {
		ICompilationUnit cu = null;
		String classnotest = null;
		
		for (int i=0;i<classesnotests.length;i++){
			try {
				
				//Concats packagename and classname
				if (classesnotests[i] instanceof ICompilationUnit){
					cu = (ICompilationUnit) classesnotests[i];
					classnotest = cu.getParent().getElementName();
					if (!classnotest.equals(""))
						classnotest = classnotest.concat(".");
					classnotest = classnotest.concat(cu.getElementName());
				}
				
				//Class is a projectmember
				if (classnotest != null && classnotest.equals(javaframe.getDeclaringTypeName()+".java")){
					TestedMethod testedmethod = new TestedMethod(javaframe.getDeclaringTypeName(), javaframe.getMethodName(), javaframe.getArgumentTypeNames());
					testedmethod.addLineNumber(javaframe.getLineNumber());
					Iterator iter = test.testedMethods.iterator();
					while (iter.hasNext()){
						TestedMethod tm = (TestedMethod) iter.next();
						if (tm.getKlass().equals(testedmethod.getKlass()) && tm.getMethod().equals(testedmethod.getMethod()) && tm.getArgs().equals(testedmethod.getArgs())){
							if (tm.linenumbers.contains(new Integer(javaframe.getLineNumber())))
								return;
							tm.addLineNumber(javaframe.getLineNumber());
							return;
						}
					}
					test.testedMethods.add(testedmethod);
				}
			} catch (DebugException e) {
				e.printStackTrace();
			} catch(Exception e1){
				e1.printStackTrace();
			}
		}
	}
	
	private boolean isDeclaringTypeInProject(IJavaElement[] allclasses, String declaringTypeName) {
		IPackageFragment [] packages;
		String fullname;
		if (allclasses[0]==null)
			return false;
		
		try {
			packages = allclasses[0].getJavaProject().getPackageFragments();
			for (int i=0; i < allclasses.length; i++){
				for (int j=0; j<packages.length;j++){
					if (packages[j].getElementName().equals(""))
						fullname = allclasses[i].getElementName();
					else
						fullname = packages[j].getElementName()+"."+allclasses[i].getElementName();
					if (fullname.equals(declaringTypeName+".java"))
						return true;
				}
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testsStarted(int, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testsStarted(int testCount, IJavaProject project) {
		testsstartet = true;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testsFinished(org.eclipse.jdt.core.IJavaProject)
	 */
	public void testsFinished(IJavaProject project) {
		testsstartet = false;
		if (JUnitPlugin.trace){
			long time;
			JUnitPlugin.endtime = Calendar.getInstance().getTimeInMillis();
			System.out.println("ENDTIME: "+JUnitPlugin.endtime);
			time = JUnitPlugin.endtime - JUnitPlugin.starttime;
			System.out.println("TIME NEEDED: "+time);
			printLinkedList();
		}
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testStarted(java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testStarted(String klass, String method, IJavaProject project) {
		tests.add(new ExtendedTestResult(project, method, klass));		
	}

	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testFailed(java.lang.String, java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testFailed(String klass, String method, String trace, IJavaProject project) {
		if (tests == null)
			return;
		ExtendedTestResult test = (ExtendedTestResult) tests.getLast();
		if (klass.equals(test.getKlass())&& method.equals(test.getMethod()))
			test.setResult(trace);
	}

	
	/* (non-Javadoc)
	 * @see org.eclipse.contribution.junit.ITestRunListener#testEnded(java.lang.String, java.lang.String, org.eclipse.jdt.core.IJavaProject)
	 */
	public void testOk(String klass, String method, IJavaProject project) {		
	}

	/**
	 * @param launch The launch to set.
	 */
	public void setLaunch(ILaunch launch) {
		this.launch = launch;
	}
	/**
	 * @param project The project to set.
	 */
	public void setProject(IJavaProject project) {
		this.project = project;
	}

	/* (non-Javadoc)
	 * @see org.eclipse.jface.operation.IRunnableWithProgress#run(org.eclipse.core.runtime.IProgressMonitor)
	 */
	public IStatus run(IProgressMonitor monitor){
		try {
			project.getProject().deleteMarkers(IBreakpoint.BREAKPOINT_MARKER, true, IResource.DEPTH_INFINITE);
			setBreakpoints(new ClassSearcher().findAllClassesExceptTestCases(project, null));
			calculatePath();
		
		} catch (DebugException e){
			//fallthrough
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}finally{
			try {
				project.getProject().deleteMarkers(IBreakpoint.BREAKPOINT_MARKER, true, IResource.DEPTH_INFINITE);
			} catch (CoreException e1) {
				e1.printStackTrace();
			}
		}
		
		return Status.OK_STATUS;
	}
	
	/**
	 * finds all methods in the testcases and sets the startbreakpoints
	 * @param elements
	 * @param markertype 
	 * @param project
	 * @throws BadLocationException
	 * @throws JavaModelException
	 * @throws CoreException
	 */
	private void setBreakpoints(IJavaElement[] elements) throws BadLocationException, JavaModelException, CoreException {
		ICompilationUnit cunit;
		IMethod [] methods;
		IDocument doc;
		IType type;
		IJavaLineBreakpoint breakpoint;
		int linenumber;
		
		for (int i=0; i<elements.length; i++){
			if (elements[i] instanceof ICompilationUnit){
				cunit = (ICompilationUnit) elements[i];
				type = cunit.findPrimaryType();
				if (cunit!=null && type != null){
					methods = type.getMethods();
					for (int j = 0;j<methods.length;j++){
						doc = new Document(cunit.getSource());
						linenumber = doc.getLineOfOffset(methods[j].getSourceRange().getOffset())+1;
						int charstart = methods[j].getSourceRange().getOffset();
						int charend = methods[j].getSourceRange().getOffset() + methods[j].getSourceRange().getLength();
						
						
						int apiLevel;
						if (project != null && "1.5".equals(project.getOptions(true).get(JavaCore.COMPILER_COMPLIANCE))) { //$NON-NLS-1$
							apiLevel= AST.JLS3;
						} else {
							apiLevel= AST.JLS2;
						}
						ASTParser parser = ASTParser.newParser(apiLevel);
						parser.setSource(doc.get().toCharArray());
						parser.setProject(project);
						CompilationUnit compilationUnit= (CompilationUnit)parser.createAST(null);
						BreakPointVerifier verifier = new BreakPointVerifier(compilationUnit, linenumber);
						compilationUnit.accept(verifier);
						int newlinenumber= verifier.getValidLocation();					
						breakpoint = JDIDebugModel.createLineBreakpoint(cunit.getUnderlyingResource(), type.getFullyQualifiedName(), newlinenumber, charstart, charend, -1, true, null);
			
						if (JUnitPlugin.trace){
							System.out.println("BREAKPOINTS: Method: "+methods[j]+" linenumber: "+linenumber+" was set.");
						}//if
					}//for
				}//if
			}//if
		}//for
	}	
}